(function() {
  var Just, Nothing, accept, definedAt, identity, lens, recurse, where, _ref;

  _ref = require('data.maybe'), Just = _ref.Just, Nothing = _ref.Nothing;

  lens = require('./Lens').of;

  identity = require('./lenses').identity;

  accept = function(predicate) {
    return lens("predicate(" + (predicate.toString()) + ")", function(ma) {
      return {
        modify: function(f) {
          return f(ma).chain(function(b) {
            if (predicate(b)) {
              return Just(b);
            } else {
              return Nothing();
            }
          });
        },
        get: function() {
          return ma.chain(function(a) {
            if (predicate(a)) {
              return Just(a);
            } else {
              return Nothing();
            }
          });
        }
      };
    });
  };

  definedAt = function(abl) {
    return function(ma) {
      return abl.runM(ma).get().isJust;
    };
  };

  where = function(predm) {
    return lens("where(" + (predm.toString()) + ")", function(ma) {
      return {
        modify: function(f) {
          if (!predm(ma)) {
            return Nothing();
          } else {
            return f(ma);
          }
        },
        get: function() {
          if (!predm(ma)) {
            return Nothing();
          } else {
            return ma;
          }
        }
      };
    });
  };

  recurse = function(lensf) {
    return lens("recurse(...)", function(ma) {
      var end, next, store;
      end = {
        modify: function(f) {
          return f(ma);
        },
        get: function() {
          return ma;
        }
      };
      if (!ma.isJust) {
        return end;
      } else {
        store = lensf().runM(ma);
        next = store.get();
        if (!next.isJust) {
          return end;
        } else {
          return {
            modify: function(f) {
              return store.modify(f);
            },
            get: function() {
              return next;
            }
          };
        }
      }
    });
  };

  module.exports = {
    accept: accept,
    definedAt: definedAt,
    where: where,
    recurse: recurse
  };

}).call(this);
